"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = createStore;

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function createStore() {
  var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  var middleware = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
  var subscribers = [];
  state = Object.assign({}, state);
  middleware = middleware.map(function (fn) {
    return fn(getState, setState);
  });

  function subscribe(subscriber) {
    subscribers.push(subscriber);
    return function () {
      unsubscribe(subscriber);
    };
  }

  function unsubscribe(subscriber) {
    var activeSubscribers = subscribers.reduce(function (out, subs) {
      if (subs === subscriber) {
        subscriber = null;
      } else {
        out.push(subs);
      }

      return out;
    }, []);
    subscribers = activeSubscribers;
  }

  function action(action) {
    return function storeAction(props, behaviors) {
      return action.apply(this, [getState, middleware, setState, props, behaviors]);
    };
  }

  function setState(updateState) {
    if (process.env.NODE_ENV === 'development') {
      console.log('%c prev state', 'color: grey', state);
    }

    state = _objectSpread({}, state, {}, updateState);

    if (process.env.NODE_ENV === 'development') {
      console.log('%c new state', 'color: maroon', state);
    }

    return subscribers.map(function (subscriber) {
      return subscriber(state);
    });
  }

  function getState() {
    return state;
  }

  return {
    getState: getState,
    action: action,
    subscribe: subscribe,
    unsubscribe: unsubscribe,
    setState: setState,
    dispatch: getState
  };
}